import grok

from zope.publisher.interfaces.browser import IBrowserRequest

from horae.core import interfaces

UNITS = ('B', 'KB', 'MB', 'GB',)


class SizeFormatter(grok.Adapter):
    """ Formats a given value as a size appended with the appropriate unit (B, KB, MB, GB)
    """
    grok.context(IBrowserRequest)
    grok.implements(interfaces.ISizeFormatter)

    def format(self, value):
        """ Formats the given value (in Bytes)
        """
        formatter = self.context.locale.numbers.getFormatter('decimal')
        unit = UNITS[0]
        i = 1
        while value > 1024 and i < len(UNITS):
            unit = UNITS[i]
            value /= 1024.0
            i += 1
        return ' '.join((formatter.format(value, '#'), unit))



import grok
from zope import schema
from zope.site.hooks import getSite

from horae.core import interfaces


class AppConfiguration(grok.Model):
    """ Holds application configuration
    """
    grok.implements(interfaces.IAppConfiguration)

    title = schema.fieldproperty.FieldProperty(interfaces.IAppConfiguration['title'])
    description = schema.fieldproperty.FieldProperty(interfaces.IAppConfiguration['description'])
    workdays = schema.fieldproperty.FieldProperty(interfaces.IAppConfiguration['workdays'])
    currency_format = schema.fieldproperty.FieldProperty(interfaces.IAppConfiguration['currency_format'])


@grok.adapter(interfaces.IAppConfigurationHolder)
@grok.implementer(interfaces.IAppConfiguration)
def app_configuration_of_holder(holder):
    if not 'app_configuration' in holder:
        holder['app_configuration'] = AppConfiguration()
    return holder['app_configuration']


class Workdays(grok.GlobalUtility):
    """ Workdays provider
    """
    grok.implements(interfaces.IWorkdays)

    _workdays = None

    def invalidate(self):
        """ Invalidates the cache of the provider
        """
        self._workdays = None

    def workdays(self):
        """ Returns a list of week days to be considered as work days

            0: Monday
            ...
            6: Sunday
        """
        if self._workdays is None:
            app = getSite()
            self._workdays = getattr(interfaces.IAppConfiguration(app), 'workdays', range(0, 6))
        return self._workdays



import grok

from z3c.objpath.interfaces import IObjectPath
from z3c.objpath import path, resolve


class ObjectPath(grok.GlobalUtility):
    grok.provides(IObjectPath)

    def path(self, obj):
        return path(grok.getSite(), obj)

    def resolve(self, path):
        return resolve(grok.getSite(), path)



from zope.schema import vocabulary

from horae.cache.vocabulary import cache_global

from horae.core.utils import getRequest


@cache_global
def weekdays_vocabulary_factory(context):
    """ A vocabulary of all days of a week registered as
        **horae.core.vocabulary.weekdays**
    """
    terms = []
    try:
        request = getRequest()
        for id, names in request.locale.dates.calendars['gregorian'].days.items():
            if names[0] is not None:
                terms.append(vocabulary.SimpleTerm(id, id, names[0]))
    except:
        pass
    return vocabulary.SimpleVocabulary(terms)
vocabulary.getVocabularyRegistry().register('horae.core.vocabulary.weekdays', weekdays_vocabulary_factory)



import grok
import transaction

from BTrees.OOBTree import OOBTree

from zope import component
from zope.event import notify
from zope.component.interfaces import IObjectEvent, ObjectEvent

from grokcore.view.publication import GrokBrowserFactory, GrokBrowserPublication

from horae.core import utils
from horae.core import interfaces


class DeferredObjectAdded(ObjectEvent):
    """ Event fired if an object has been added to a deferred utility
    """
    grok.implements(interfaces.IDeferredObjectAdded)


class BaseQueue(object):
    """ Abstract implementation of a queue
    """

    index_factory = OOBTree
    event_factory = DeferredObjectAdded

    @property
    def _queue(self):
        try:
            request = utils.getRequest(check=False)
            if not hasattr(request, self._queue_attr):
                setattr(request, self._queue_attr, [])
            return getattr(request, self._queue_attr)
        except:
            return None

    @property
    def _index(self):
        try:
            request = utils.getRequest(check=False)
            if not hasattr(request, self._index_attr):
                setattr(request, self._index_attr, self.index_factory())
            return getattr(request, self._index_attr)
        except:
            return None

    def _key(self, obj):
        return obj

    def action(self, obj):
        raise NotImplementedError(u'concrete class must implement action()')

    def add(self, obj):
        """ Adds an object to the queue
        """
        if self._queue is not None and self._index is not None:
            key = self._key(obj)
            if key in self._index:
                if self._index[key] in self._queue:
                    self._queue.remove(self._index[key])
            self._queue.append(obj)
            self._index[key] = obj
            notify(self.event_factory(obj))
        else:
            self.action(obj)

    def clear(self):
        """ Clear the queue
        """
        try:
            request = utils.getRequest(check=False)
            setattr(request, self._queue_attr, [])
            setattr(request, self._index_attr, self.index_factory())
        except:
            pass

    def __call__(self):
        """ Processes the queue
        """
        if self._queue is None:
            return
        while len(self._queue):
            obj = self._queue.pop(0)
            self.action(obj)
            transaction.commit()


class DeferredNotifierObjectAdded(DeferredObjectAdded):
    """ Event fired if an object has been added to the deferred notifier utility
    """
    grok.implements(interfaces.IDeferredNotifierObjectAdded)


class DeferredNotifier(BaseQueue, grok.GlobalUtility):
    """ Utility handling deferred notification of events
    """
    grok.implements(interfaces.IDeferredNotifier)

    _queue_attr = '__horae_core_notifyqueue'
    _index_attr = '__horae_core_notifyindex'
    event_factory = DeferredNotifierObjectAdded

    def _key(self, obj):
        if IObjectEvent.providedBy(obj):
            if interfaces.IIntId.providedBy(obj.object):
                return (obj.__class__.__name__, obj.object.__class__.__name__, obj.object.id)
            return (obj.__class__.__name__, obj.object)
        return obj

    def action(self, obj):
        notify(obj)

    def notify_objects(self, interface):
        """ Notify object events of objects implementing the given interface
        """
        if self._queue is None:
            return
        index = 0
        while len(self._queue) > index:
            if (not IObjectEvent.providedBy(self._queue[index]) or
                not interface.providedBy(self._queue[index].object)):
                index += 1
                continue
            obj = self._queue.pop(index)
            self.action(obj)
            transaction.commit()


class BrowserPublication(GrokBrowserPublication):
    def callObject(self, request, ob):
        result = super(BrowserPublication, self).callObject(request, ob)
        for utility in component.getAllUtilitiesRegisteredFor(interfaces.IDeferred):
            utility()
        return result


class BrowserFactory(GrokBrowserFactory):
    def __call__(self):
        request_class, publication = super(BrowserFactory, self).__call__()
        return request_class, BrowserPublication



import grok
import logging

from zope import component

from horae.core import interfaces

logger = logging.getLogger('horae.core.upgrade')


class ApplicationUpgrader(grok.Adapter):
    """ An adapter upgrading an application
    """
    grok.context(interfaces.IHorae)
    grok.implements(interfaces.IApplicationUpgrader)

    def log(self, *msgs):
        for msg in msgs:
            self.results.append(msg)
            logger.info(msg)

    def upgrade(self):
        """ Runs all available registered upgrade steps
        """
        steps = component.getAdapters((self.context,), interfaces.IApplicationUpgrade)
        self.results = []
        self.log('*' * 50)
        self.log('STARTING APPLICATION UPGRADE %s' % self.context.__name__)
        for name, step in steps:
            if not step.available():
                continue
            self.log('-' * 50)
            self.log('starting %s' % name)
            self.log('-' * 50)
            try:
                self.log(*step.upgrade())
            except:
                self.log('-' * 50)
                self.log('%s failed' % name)
                self.log('-' * 50)
                continue
            self.log('-' * 50)
            self.log('finished %s' % name)
            self.log('-' * 50)
        self.log('FINISHED APPLICATION UPGRADE')
        self.log('*' * 50)
        return self.results



# -*- coding: utf-8 -*-
from zope import interface
from zope import schema
from zope.component.interfaces import IObjectEvent

from horae.core import _


class IHorae(interface.Interface):
    """ Marker interface for Horae applications
    """


class IApplicationUpgrader(interface.Interface):
    """ An adapter upgrading an application
    """

    def upgrade():
        """ Runs all registered and available :py:class:`IApplicationUpgrade`
        """


class IApplicationUpgrade(interface.Interface):
    """ An application upgrade step
    """

    def available():
        """ Whether this step is available
        """

    def upgrade():
        """ Runs the upgrade and returns a list of strings containing the result
        """


class ITextId(interface.Interface):
    """ An object having a text ID
    """

    id = schema.ASCIILine(
        title=_(u'ID'),
        required=True
    )


class IIntId(interface.Interface):
    """ An object having an integer ID
    """

    id = schema.Int(
        title=_(u'ID'),
        required=True
    )


class ITextIdManager(interface.Interface):
    """ Handles text IDs
    """

    def normalize(name):
        """ Returns a normalized string usable in URLs based on the ``name`` provided
        """

    def idFromName(container, name):
        """ Returns a valid ID for a new object to be added to the ``container``
            from the ``name`` provided
        """


class IIntIdManager(interface.Interface):
    """ Handles integer IDs
    """

    def nextId(key):
        """ Returns the next ID for the given ``key``
        """


class IContainer(interface.Interface):
    """ A container for objects implementing IIntId
    """

    def add_object(obj):
        """ Adds a new object and returns the generated ID
        """

    def get_object(id):
        """ Returns the specified object
        """

    def del_object(id):
        """ Deletes the specified object
        """

    def objects():
        """ Iterator over the contained objects
        """

    def last():
        """ Returns the last added object
        """


class IPriced(interface.Interface):
    """ A priced object
    """

    def price():
        """ The price of this object
        """


class IWorkdays(interface.Interface):
    """ Workdays provider
    """

    def invalidate():
        """ Invalidates the cache of the provider
        """

    def workdays():
        """ Returns a list of week days to be considered as work days

            0: Monday
            ...
            6: Sunday
        """


class IAppConfiguration(interface.Interface):
    """ Holds application configuration
    """

    title = schema.TextLine(
        title=_(u'Title'),
        description=_(u'The title of the application'),
        default=u'Horae',
        required=True
    )

    description = schema.TextLine(
        title=_(u'Description'),
        description=_(u'The description of the application'),
        default=u'Next generation resource planning'
    )

    workdays = schema.Set(
        title=_(u'Work days'),
        required=True,
        value_type=schema.Choice(
            vocabulary='horae.core.vocabulary.weekdays'
        ),
        default=set(range(1, 6))
    )

    currency_format = schema.TextLine(
        title=_(u'Currency format'),
        required=True,
        default=u'%s'
    )

    @interface.invariant
    def currencyFormatContainsPlaceholder(app_configuration):
        if not '%s' in app_configuration.currency_format:
            raise interface.Invalid(_(u'The currency format is invalid, please insert «%s» as a placeholder for the value'))


class IAppConfigurationHolder(interface.Interface):
    """ Marker interface for objects adaptable to :py:class:`IAppConfiguration`
    """


class ICurrencyFormatter(interface.Interface):
    """ Formats a given value using the currency format specified in the
        :py:class:`IAppConfiguration`
    """

    def format(value, symbol=True):
        """ Formats the given value
        """


class ISizeFormatter(interface.Interface):
    """ Formats a given value as a size pretended with the appropriate unit (KB, MB, GB)
    """

    def format(value):
        """ Formats the given value (in Bytes)
        """


class IDeferred(interface.Interface):
    """ Utility called after page processing
    """

    def add(obj):
        """ Adds an object to the queue
        """

    def __call__():
        """ Processes the queue
        """

    def clear():
        """ Clear the queue
        """


class IDeferredObjectAdded(IObjectEvent):
    """ Event fired if an object has been added to a :py:class:`IDeferred`
    """


class IDeferredNotifier(IDeferred):
    """ Utility handling deferred notification of events
    """

    def notify_objects(interface):
        """ Notify object events of objects implementing the given interface
        """


class IDeferredNotifierObjectAdded(IDeferredObjectAdded):
    """ Event fired if an object has been added to the :py:class:`IDeferredNotifier`
    """



import grok

from zope import component

from horae.core import interfaces


class Container(grok.OrderedContainer):
    """ A container for objects implementing IIntId ot ITextId (and having a name attribute)
    """
    grok.implements(interfaces.IContainer)
    grok.baseclass()

    def id_key(self):
        return 'object'

    def add_object(self, obj):
        """ Adds a new object and returns the generated id
        """
        if not hasattr(obj, 'id') or obj.id is None:
            if interfaces.ITextId.providedBy(obj):
                obj.id = component.getUtility(interfaces.ITextIdManager).idFromName(self, obj.name)
            else:
                obj.id = component.getUtility(interfaces.IIntIdManager).nextId(self.id_key())
        self._last = obj
        self[str(obj.id)] = obj
        return obj.id

    def get_object(self, id):
        """ Returns the specified object
        """
        return self[str(id)]

    def del_object(self, id):
        """ Deletes the specified object
        """
        del self[str(id)]

    def objects(self):
        """ Iterator over the contained objects
        """
        return self.values()

    def last(self):
        """ Returns the last added object
        """
        return getattr(self, '_last', None)



import math
from datetime import date, timedelta, datetime

from BTrees.IOBTree import IOBTree
from BTrees.OIBTree import OIBTree

from zope import security
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.i18n import translate

from horae.core import _

_marker = object()


def getRequest(default=_marker, check=True):
    """ Returns the current request implementing
        :py:class:`zope.publisher.interfaces.browser.IBrowserRequest`
        raises a RuntimeError if no request is found and no default
        value is provided
    """
    i = security.management.getInteraction() # raises NoInteraction
    for p in i.participations:
        if not check or IBrowserRequest.providedBy(p):
            return p
    if default is not _marker:
        return default
    raise RuntimeError('Could not find current request.')


def getPrincipal():
    """ Returns the currently used principal or None
    """
    try:
        return getRequest().principal
    except:
        return None


def filterItems(items, permission='horae.View'):
    """ Filters a list of objects by permission
    """
    filtered = []
    for item in items:
        if security.checkPermission(permission, item):
            filtered.append(item)
    return filtered

_cache_id = OIBTree()
_cache = IOBTree()


def findParentByInterface(context, interface, skip=0):
    """ Returns the first found parent of the context implementing the given
        interface, optionally skipping a defined number of matching parents
    """
    id = _cache_id.get((context, interface, skip), _marker)
    if id is _marker:
        id = len(_cache_id)
        _cache_id[(context, interface, skip)] = id
    cached = _cache.get(id, _marker)
    if cached is not _marker:
        return cached
    while context is not None and (not interface.providedBy(context) or skip > 0):
        if interface.providedBy(context):
            skip -= 1
        context = getattr(context, '__parent__', None)
    if interface.providedBy(context):
        _cache[id] = context
        return context
    _cache[id] = None
    return None


def formatDateTime(value, request, format=('dateTime', 'short'), html=True):
    """ Formats the datetime ``value`` using the locale provided by the ``request``
        and the ``format`` provided. HTML output may be disabled by the ``html``
        argument.
    """
    formatter = request.locale.dates.getFormatter(*format)
    if value is None:
        return u''
    formatted = None
    if format[0].startswith('date'):
        today = date.today()
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        d = value.date() if isinstance(value, datetime) else value
        if d == today:
            formatted = translate(_(u'Today'), context=request)
        elif d == yesterday:
            formatted = translate(_(u'Yesterday'), context=request)
        elif d == tomorrow:
            formatted = translate(_(u'Tomorrow'), context=request)
        if formatted is not None and format[0].lower().endswith('time'):
            timeformatter = request.locale.dates.getFormatter('time', format[1])
            formatted += ' ' + timeformatter.format(value)
    if formatted is None:
        formatted = formatter.format(value)
    if not html:
        return formatted
    formatter = request.locale.dates.getFormatter('dateTime', 'long')
    return '<time datetime="%s" title="%s">%s</time>' % (value.isoformat(), formatter.format(value), formatted)


def formatDateTimeRange(start, end, request, format=('dateTime', 'short'), html=True):
    """ Formats the datetime range (``start``, ``end``) using the locale provided
        by the ``request`` and the ``format`` provided. HTML output may be disabled
        by the ``html`` argument.
    """
    s = start.date() if isinstance(start, datetime) else start
    e = end.date() if isinstance(end, datetime) else end
    start = formatDateTime(start, request, format, html)
    if s == e:
        end = formatDateTime(end, request, ('time', format[1]), html)
    else:
        end = formatDateTime(end, request, format, html)
    return '%s - %s' % (start, end)


def formatHours(value, request):
    """ Formats the ``value`` as a decimal using the locale provided by the ``request``
    """
    if value is None:
        return u''
    formatter = request.locale.numbers.getFormatter('decimal')
    if value == math.floor(value):
        return formatter.format(value, '#,##0;-#,##0')
    return formatter.format(value, '#,##0.00;-#,##0.00')



import grok
import re
from threading import Lock

from zope.site.hooks import getSite
from zope.annotation.interfaces import IAnnotations
from persistent.dict import PersistentDict

from horae.core import interfaces

INTID_ANNOTATIONS_KEY = 'horae.core.intid'
int_id_lock = Lock()


class IntIdManager(grok.LocalUtility):
    """ Handles integer IDs
    """
    grok.implements(interfaces.IIntIdManager)

    def __init__(self):
        self._storage = None

    @property
    def storage(self):
        if self._storage is None:
            storage = IAnnotations(getSite())
            if not INTID_ANNOTATIONS_KEY in storage:
                storage[INTID_ANNOTATIONS_KEY] = PersistentDict()
            self._storage = storage[INTID_ANNOTATIONS_KEY]
        return self._storage

    def nextId(self, key):
        """ Returns the next ID for the given ``key``
        """
        int_id_lock.acquire()
        if not key in self.storage:
            self.storage[key] = 0
        self.storage[key] += 1
        id = self.storage[key]
        int_id_lock.release()
        return id


class TextIdManager(grok.GlobalUtility):
    """ Handles text IDs
    """
    grok.implements(interfaces.ITextIdManager)
    re_ws = re.compile(r'\s+')
    re_chars = re.compile(r'[^a-zA-Z0-9-_\.]')

    def normalize(self, name):
        """ Returns a normalized string usable in URLs based on the ``name`` provided
        """
        return self.re_chars.sub('', self.re_ws.sub('_', name)).lower()

    def idFromName(self, container, name):
        """ Returns a valid ID for a new object to be added to the ``container``
            from the ``name`` provided
        """
        id = self.normalize(name)
        if id in container:
            i = 2
            while '%s-%s' % (id, i) in container:
                i += 1
            id = '%s-%s' % (id, i)
        return str(id)



import grok
import transaction

from z3c.taskqueue import service, interfaces, startup
import z3c.taskqueue

from zope.site.hooks import getSite, setSite

from zope import component, interface
from zope.event import notify
from zope.security import management
from zope.processlifetime import IDatabaseOpenedWithRoot
from zope.app.publication.zopepublication import ZopePublication

from horae.core.interfaces import IHorae

task_service = None


class IHoraeServiceInitialized(interface.Interface):
    """ Notifies the initialization of the horae task service
    """

    service = interface.Attribute('The task service instance')


class HoraeServiceInitialized(object):
    interface.implements(IHoraeServiceInitialized)

    def __init__(self, service):
        self.service = service


@grok.subscribe(IDatabaseOpenedWithRoot)
def databaseOpened(event, productName='z3c.taskqueue'):
    global task_service
    db = event.database
    connection = db.open()
    root = connection.root()
    root_folder = root.get(ZopePublication.root_name, None)
    if not 'HoraeTaskService' in root_folder:
        root_folder['HoraeTaskService'] = service.TaskService()
        root_folder['HoraeTaskService'].__name__ = 'HoraeTaskService'
        notify(HoraeServiceInitialized(root_folder['HoraeTaskService']))
    task_service = root_folder['HoraeTaskService']
    task_service.processorArguments = {}
    component.provideUtility(task_service, interfaces.ITaskService, name='HoraeTaskService')
    transaction.commit()

    # START copy from z3c.taskqueue.startup.databaseOpened
    startup.log.info('handling event IDatabaseOpenedEvent')

    startup.storeDBReferenceOnDBOpened(event)

    from zope.app.appsetup.product import getProductConfiguration
    configuration = getProductConfiguration(productName)
    startSpecifications = startup.getStartSpecifications(configuration)

    for sitesIdentifier, servicesIdentifier in startSpecifications:
        startedAnything = False
        sites = startup.getSites(sitesIdentifier, root_folder)
        for site in sites:
            if servicesIdentifier == '*':
                started = startup.startAllServices(site, root_folder)
            else:
                started = startup.startOneService(site, servicesIdentifier)
            startedAnything = startedAnything or started

        if sitesIdentifier == "*" and not startedAnything:
            msg = 'no services started by directive *@%s'
            startup.log.warn(msg % servicesIdentifier)
    # END copy from z3c.taskqueue.startup.databaseOpened


def sites():
    db = z3c.taskqueue.GLOBALDB
    connection = db.open()
    root = connection.root()
    root_folder = root.get(ZopePublication.root_name, None)
    interaction = management.getInteraction()
    participation = interaction.participations[0]
    participation.setPrincipal(management.system_user)
    transaction.commit()
    _site = getSite()
    for site in root_folder.values():
        if IHorae.providedBy(site):
            setSite(site)
            yield site
    setSite(_site)
    transaction.commit()
    connection.close()



import grok

from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserRequest

from horae.core import interfaces
from horae.core import utils


class CurrencyFormatter(grok.MultiAdapter):
    """ Formats a given value using the currency format specified in the application configuration
    """
    grok.adapts(Interface, IBrowserRequest)
    grok.implements(interfaces.ICurrencyFormatter)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def format(self, value, symbol=True):
        """ Formats the given value
        """
        formatter = self.request.locale.numbers.getFormatter('decimal')
        value = formatter.format(value, '#,##0.00;-#,##0.00')
        parent = utils.findParentByInterface(self.context, interfaces.IAppConfigurationHolder)
        if not symbol or parent is None:
            return value
        try:
            return interfaces.IAppConfiguration(parent).currency_format % value
        except:
            return value
